/* $NetBSD: drm_atomic.h,v 1.7 2008/06/03 20:35:01 dyoung Exp $ */

/*-
 * Copyright (c) 2008 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Andrew Doran, Blair Sadewitz, et. al.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * XXXbjs This code is essentially a frob of the atomic operations provided by
 * 	  the Linux kernel at the time the DRM was ported.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: drm_atomic.h,v 1.7 2008/06/03 20:35:01 dyoung Exp $");
#ifdef __NetBSD__
#include <sys/cdefs.h>
#include <sys/atomic.h>
#elif defined(__OpenBSD__)
#include <machine/atomic.h>
#endif

typedef uint32_t atomic_t;

#ifdef __FreeBSD__
#define	atomic_set(p, v)		(*((volatile uint32_t *)p) = (v))
#define	atomic_read(p)			(*((volatile uint32_t *)p))
#define atomic_inc(p)			atomic_add_int(p, 1)
#define atomic_dec(p)			atomic_subtract_int(p, 1)
#define atomic_add(n, p)		atomic_add_int(p, n)
#define atomic_sub(n, p)		atomic_subtract_int(p, n)
#elif defined(__NetBSD__)
#define	atomic_set(p, v)		(*((volatile uint32_t *)p) = (v))
#define	atomic_read(p)			(*((volatile uint32_t *)p))
#define	atomic_inc(p)			atomic_inc_uint(p)
#define	atomic_dec(p)			atomic_dec_uint(p)
#define	atomic_add(v, p)		atomic_add_int(p, v)
#define	atomic_sub(v, p)		atomic_add_int(p, -(v))
#define	atomic_set_int(p, bits)		atomic_or_uint(p, bits)
#define	atomic_clear_int(p, bits)	atomic_and_uint(p, ~(bits))
#define	atomic_cmpset_int(p, o, n)					\
			((old == atomic_cas_uint(p, o, n)) ? 1 : 0)
#else
/* XXX certain versions of GCC on certain platforms have builtin atomic ops
 *     If this is not available, platforms which support restartable atomic
 */    sequences should use those if possible.
#define	atomic_set(p, v)		(*((volatile uint32_t *)p) = (v))
#define	atomic_read(p)			(*((volatile uint32_t *)p))
#define atomic_inc(p)			(*((volatile uint32_t *)p) += 1)
#define atomic_dec(p)			(*((volatile uint32_t *)p) -= 1)
#define atomic_add(n, p)		(*((volatile uint32_t *)p) += (n))
#define atomic_sub(n, p)		(*((volatile uint32_t *)p) -= (n))
#define atomic_add_int(p, v)      	(*((volatile uint32_t *)p) += v)
#define atomic_subtract_int(p, v) 	(*((volatile uint32_t *)p) -= v)
#ifdef __OpenBSD__
#define atomic_set_int(p, bits)		atomic_setbits_int(p,bits)
#define atomic_clear_int(p, bits)	atomic_clearbits_int(p,bits)
#else
#define atomic_set_int(p, bits)   	(*((volatile uint32_t *)p) |= (bits))
#define atomic_clear_int(p, bits) 	(*((volatile uint32_t *)p) &= ~(bits))
#endif
#endif /* !__FreeBSD__ */

#if !defined(__NetBSD__)
#if !defined(__FreeBSD_version) || (__FreeBSD_version < 500000)
#if defined(__i386__) || defined(__amd64__)
/* The extra atomic functions from 5.0 haven't been merged to 4.x */
static __inline int
atomic_cmpset_int(volatile u_int *dst, u_int exp, u_int src)
{
	int res = exp;

	__asm __volatile (
	"	lock ;			"
	"	cmpxchgl %1,%2 ;	"
	"       setz	%%al ;		"
	"	movzbl	%%al,%0 ;	"
	"1:				"
	"# atomic_cmpset_int"
	: "+a" (res)			/* 0 (result) */
	: "r" (src),			/* 1 */
	  "m" (*(dst))			/* 2 */
	: "memory");				 

	return (res);
}
#else /* __i386__ */
static __inline int
atomic_cmpset_int(volatile int *dst, int old, int new)
{
	int s = splhigh();
	if (*dst==old) {
		*dst = new;
		splx(s);
		return 1;
	}
	splx(s);
	return 0;
}
#endif /* !__i386__ */
#endif /* !__FreeBSD_version || __FreeBSD_version < 500000 */
#endif /* !__NetBSD__ */

#define	set_bit(b, p)	\
	atomic_set_int(((volatile uint32_t *)(volatile void *)p) + (b >> 5), \
			(1 << (b & 0x1f)))

#define	clear_bit(b, p)  \
	atomic_clear_int(((volatile uint32_t *)(volatile void *)p) + (b >> 5),\
			(1 << (b & 0x1f)))

#define	test_bit(b, p)    \
	(((volatile uint32_t *)(volatile void *)p)[b >> 5] & (1 << (b & 0x1f)))

static inline uint32_t 
test_and_set_bit(int b, volatile void *p)
{
	volatile uint32_t *val;
	uint32_t mask, old;

	val = (volatile uint32_t *)p;
	mask = 1 << b;

	do {
		old = *val;
		if ((old & mask) != 0)
			break;
	} while (!atomic_cmpset_int(val, old, old | mask));

	return old & mask;
}
